#ifndef cg_h
#define cg_h
#include "CGsystem.h"
#include "math.h"
#include "cublas.h"

template <class T>
T cg(CGsystem<T> *A,const T tol, const int maxIt)
{
	//rnorm is the square of the 2-norm of the residual.
	T alpha, beta, rnorm, oldnorm, previousnorm, bestNorm;
	unsigned int iterationsSoFar = 0;
	int maxIterations = maxIt;
	T pTAp;
	T *x, *r, *p, *Ap, *b, *bestx;
	//initialize variables:
	T **vecArray = A->initVectors(&x, &b, 4);
	Ap = vecArray[0];
	r = vecArray[1];
	p = vecArray[2];
	bestx = vecArray[3];
	A->scale(0.0, x);					// x = 0
	A->scale(0.0, r);
	A->axpy(1, b, r);					// r = b;
	rnorm = A->vectorVectorProduct(r,r);
	T t = tol*tol*rnorm;
	T adjustment = 0.8;
	T etta = t * adjustment;
	previousnorm = rnorm;
	bestNorm = rnorm;
	do{
		A->scale(0.0, p);
		A->axpy(1, r, p);							// p = r
		unsigned int k;
		for (k=0; (k < A->getN()) && (k + iterationsSoFar < maxIterations) && (rnorm > etta); ++k){
			A->matrixVectorProduct(p, Ap); //Ap = A*p
			pTAp = A->vectorVectorProduct(p, Ap);
//			pTAp = cublasSdot (A->getN(), (float *)p, 1, (float*)Ap, 1);

			//if pTAp == 0, we have a lucky breakdown, if pTAp < 0 something is wrong...
			if (pTAp <= (T)0.0) break;
			alpha = rnorm / pTAp;
			if (alpha == (T)0.0) break;
			//saxpy calculating x += alpha*p and r -= alpha*Ap:
			A->axpy(-alpha, Ap, r);
			oldnorm = rnorm;
			rnorm = A->vectorVectorProduct(r,r);
			if (rnorm > oldnorm && oldnorm < bestNorm){
				A->copy(x, &bestx);
				bestNorm = oldnorm;
			}
			A->axpy(alpha, p, x);
//			rnorm = cublasSdot (A->getN(), (float *)r, 1, (float*)r, 1);
			beta = rnorm / oldnorm;
			//saxpy for p = beta * p + r.
			A->scale(beta, p);
			A->axpy(1.0, r, p);
			if (k % 10000 == 0){
/*				if (sizeof(T) == sizeof(float)){
					T crnorm = cublasSdot (A->getN(), (float *)r, 1, (float*)r, 1);
					printf("difference %g\n", crnorm - rnorm);
				}
				if (sizeof(T) == sizeof(double)){
					T crnorm = cublasDdot (A->getN(), (double *)r, 1, (double*)r, 1);
					printf("difference %g\n", crnorm - rnorm);
				}
*/				writeToScreen1("k = %i\t", k);
				writeToScreen1("norm: %g\t", sqrt(oldnorm));
				writeToScreen1("best norm so far: %g\t", sqrt(bestNorm));
				writeToScreen1("desired norm: %g\n", sqrt(t));
			}
		}
		iterationsSoFar += k;
		if (rnorm > bestNorm){
			A->copy(bestx, &x);
			rnorm = bestNorm;
		}
		/* check if the rnorm has lost too much accuracy due to round-off errors */
		A->matrixVectorProduct(x, p); 	// p = Ax
		A->scale(0.0, r);						// r = 0
		A->axpy(1, b, r);						// r = b
		A->axpy(-1, p, r);						// r = b - Ax
		oldnorm = rnorm;
		rnorm = A->vectorVectorProduct(r, r);
		writeToScreen1("The difference between calculated and real norm is: %g\n", (rnorm > oldnorm) ? sqrt(rnorm - oldnorm) : sqrt(oldnorm - rnorm));
		if (pTAp == 0) return sqrt(rnorm);
		if (rnorm > t && rnorm >= previousnorm){
			writeToScreen("Unable to convert to the desired precition, and restart led to the same result\n");
			return sqrt(rnorm);
		}
		if (iterationsSoFar >= maxIterations){
			writeToScreen("Unable to convert to the desired precition, \nbecause the maximum number of iterations was reached\n");
			writeToScreen1("norm: %g\n", sqrt(rnorm));
			writeToScreen1("number of iterations: %i\n", iterationsSoFar);
			return sqrt(rnorm);
		}
		if (alpha == (T)0.0 || beta == (T)0.0 || pTAp < (T)0.0){
			writeToScreen("Unable to convert to the desired precition, because the iteration stagnated: ");
			if (alpha == (T)0.0) writeToScreen("alpha became zero\n");
			else if (pTAp < (T)0.0) writeToScreen("pTAp < 0\n");
			else	writeToScreen("beta became zero\n");
			return sqrt(rnorm);
		}
		if (rnorm > t && iterationsSoFar != maxIterations) writeToScreen("Restarting...\n");
		previousnorm = rnorm;
	}while(rnorm > t && pTAp > 0);
	
	writeToScreen1("Solved the system by %i iterations\n", iterationsSoFar);
	return sqrt(rnorm);
}

#endif
