#include "mexStarter.h"
/*
 * mexStarter.cu
 * Starts a conjugate gradient solver on the GPU
 * solving either two point flux approximation
 * or mimetic discretization of a grid.
 * 
 * It can also solve a matrix-vector system Ax = b
 * if A and b are given as input.
 */

/*The gateway function from MATLAB */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if(nrhs < 3) mexErrMsgIdAndTxt("MyToolbox:mexMaster:nrhs","Too few input variables (3 for full matrix, 5 for sparse matrix or 6 for grid).\n");
	if(nrhs > 6) mexErrMsgIdAndTxt("MyToolbox:mexMaster:nrhs","Too many input variables (3 for full matrix, 5 for sparse matrix or 6 for grid).\n");
	if(nlhs == 0) return;
	if(nlhs != 2) mexErrMsgIdAndTxt("MyToolbox:mexMaster:nrhs","Wrong number of output variables (solution vector and norm-value should be handled).\n");
	
	if(mxIsDouble(prhs[0]) && mxIsDouble(prhs[1]) && mxIsDouble(prhs[2])
			&& !mxIsComplex(prhs[0]) && !mxIsComplex(prhs[1]) && !mxIsComplex(prhs[2])){
		if (nrhs <= 5){
			/* import from sparse or full matrix */
			mwSize n =mxGetM(prhs[1]);
			plhs[0] = mxCreateDoubleMatrix(n, 1, mxREAL);
			double norm = importMatrix<double>((double *)mxGetData(plhs[0]), nrhs, prhs);
			plhs[1] = mxCreateDoubleScalar(norm);
		} else { //nrhs == 6
			if (!mxIsDouble(prhs[2]))
				mexErrMsgIdAndTxt("MyToolbox:mexMaster:prhs","A, b and tol must all be of double precition for double precition calculation,\nfor single A is double, b and tol is single.\n");
			mwSize n = mxGetM(prhs[2]);
			plhs[0] = mxCreateDoubleMatrix(n, 1, mxREAL);
			double norm = importGrid<double>((double *)mxGetData(plhs[0]), prhs);
			plhs[1] = mxCreateDoubleScalar(norm);
		}
	} else {
		if(mxIsDouble(prhs[0]) && mxIsSingle(prhs[1]) && mxIsSingle(prhs[2])
				&& !mxIsComplex(prhs[0]) && !mxIsComplex(prhs[1]) && !mxIsComplex(prhs[2])){
			if (nrhs <= 5){
				/* import from sparse or full matrix */
				mwSize *dims = new mwSize[2];
				dims[0] = mxGetM(prhs[1]);
				dims[1] = 1;
				plhs[0] = mxCreateNumericArray(2, dims, mxSINGLE_CLASS, mxREAL);
				delete[] dims;
				float norm = importMatrix<float>((float *)mxGetData(plhs[0]), nrhs, prhs);
				plhs[1] = mxCreateDoubleScalar(norm);
			} else {
				/* nrhs == 6 import from permeability grid */
				if (!mxIsSingle(prhs[2]))
					mexErrMsgIdAndTxt("MyToolbox:mexMaster:prhs","A, b and tol must all be of double precition for double precition calculation,\nfor single A is double, b and tol is single.\n");
				mwSize *dims = new mwSize[2];
				dims[0] = mxGetM(prhs[2]);
				dims[1] = 1;
				plhs[0] = mxCreateNumericArray(2, dims, mxSINGLE_CLASS, mxREAL);
				delete[] dims;
				float norm = importGrid<float>((float *)mxGetData(plhs[0]),prhs);
				plhs[1] = mxCreateDoubleScalar(norm);
			}
		} else {
			mexErrMsgIdAndTxt("MyToolbox:mexMaster:prhs","A, b and tol must all be of double precition for double precition calculation,\nfor single A is double, b and tol is single.\n");
		}
	}
}

