#ifndef rcg_h
#define rcg_h
#include "CGsystem.h"
#include "math.h"
#include "SysHYB.cu.h"
//#include "cublas.h"
#include <MyProject_kernel.cu>

template <class T>
T rcg(CGsystem<T> *A, const T tol, const T tol2, const int maxIt, const T threshold)
{
	//rnorm is the square of the 2-norm of the residual.
	T oldNorm, previousNorm, rNormHigh, scaleFactor;
	T rNormLow, bestNorm, pTAp, eta, alpha, beta;
	unsigned int iterationsSoFar = 0;
	int maxIterations = maxIt;
	T * xHigh, *b;
	//initialize variables:
	T **vecArray = A->initVectors(&xHigh, &b, 7);
	T *rHigh	= vecArray[0];
	T *p			= vecArray[1];
	T *rLow	= vecArray[2];
	T *xLow	= vecArray[3];
	T *Ap		= vecArray[4];
	T *bestx	= vecArray[5];
	T *bestxHigh = vecArray[6];
	
	A->scale(0.0, xHigh);						// x = 0
	A->copy(b, &rHigh);							// r = b - Ax;
	rNormHigh = A->vectorVectorProduct(rHigh, rHigh);
	
	T endNorm = rNormHigh*tol*tol;
	do{
		previousNorm = rNormHigh;
		A->copy(rHigh, &rLow);
		scaleFactor = sqrt(rNormHigh);
		A->scale(1.0 / scaleFactor, rLow);
		A->copy(rLow, &p);
		A->scale(0.0, xLow);
		rNormLow = A->vectorVectorProduct(rLow, rLow);
		bestNorm = rNormLow;
		if (threshold * endNorm / rNormHigh > tol2*tol2)
			eta = threshold * endNorm / rNormHigh;
		else
			eta = rNormLow*tol2*tol2;
		unsigned int k;
		writeToScreen1("eta: %g\n", sqrt(eta));
		for (k=0; (k < A->getN()) && (k + iterationsSoFar < maxIterations) && (rNormLow > eta); ++k){
			A->matrixVectorProduct(p, Ap); //Ap = A*p
			pTAp = A->vectorVectorProduct(p, Ap);
			//if pTAp == 0, we have a lucky breakdown, if pTAp < 0 something is wrong...
			if (pTAp <= (T)0.0) break;
			alpha = rNormLow / pTAp;
			if (alpha == (T)0.0) break;
			//saxpy calculating x += alpha*p and r -= alpha*Ap:
			A->axpy(-alpha, Ap, rLow);
			oldNorm = rNormLow;
			rNormLow = A->vectorVectorProduct(rLow,rLow);
			if (rNormLow > oldNorm && oldNorm < bestNorm){
				A->copy(xLow, &bestx);
				bestNorm = oldNorm;
			}
			A->axpy(alpha, p, xLow);
			beta = rNormLow / oldNorm;
			//saxpy for p = beta * p + r.
			A->scale(beta, p);
			A->axpy(1.0, rLow, p);
			if (k % 10000 == 0){
/*				if (sizeof(T) == sizeof(float)){
					T crnorm = cublasSdot (A->getN(), (float *)r, 1, (float*)r, 1);
					printf("difference %g\n", crnorm - rnorm);
				}
				if (sizeof(T) == sizeof(double)){
					T crnorm = cublasDdot (A->getN(), (double *)r, 1, (double*)r, 1);
					printf("difference %g\n", crnorm - rnorm);
				}
*/				writeToScreen1("IterationsSoFar = %i\t", k + iterationsSoFar);
				writeToScreen1("norm: %g\t", sqrt(oldNorm));
				writeToScreen1("best norm so far: %g\t", sqrt(bestNorm));
				writeToScreen1("desired norm: %g\n", sqrt(eta));
			}
		}
		iterationsSoFar += k;
		if (rNormLow > bestNorm){
			A->copy(bestx, &xLow);
			rNormLow = bestNorm;
		}
		/* Update x by the xLow: xHigh = xHigh + ||xHigh||*xLow */
		A->copy(xHigh, &bestxHigh);
		A->axpy(scaleFactor, xLow, xHigh);
		
		A->matrixVectorProduct(xHigh, p); 	// p = Ax
		A->copy(b, &rHigh);						// r = b
		A->axpy(-1, p, rHigh);						// r = b - Ax
		rNormHigh = A->vectorVectorProduct(rHigh, rHigh);
		writeToScreen1("The new norm is: %g\n", sqrt(rNormHigh));
		if (pTAp == 0) return sqrt(rNormHigh);
		if (rNormHigh > endNorm && rNormHigh >= previousNorm){
			writeToScreen("Restart led to the same or worse result\n");
			A->copy(bestxHigh, &xHigh);
			return sqrt(previousNorm);
		}
		if (iterationsSoFar >= maxIterations){
			writeToScreen("Unable to convert to the desired precition, \nbecause the maximum number of iterations was reached\n");
			writeToScreen1("norm: %g\n", sqrt(rNormHigh));
			writeToScreen1("number of iterations: %i\n", iterationsSoFar);
			return sqrt(rNormHigh);
		}
		if (alpha == (T)0.0 || beta == (T)0.0 || pTAp < (T)0.0){
			
			writeToScreen("Unable to convert to the desired precition, because the iteration stagnated: ");
			if (alpha == (T)0.0) writeToScreen("alpha became zero\n");
			else if (pTAp < (T)0.0) writeToScreen1("pTAp = %g\n", pTAp);
			else	writeToScreen("beta became zero\n");
			return sqrt(rNormHigh);
		}
		if (rNormHigh > endNorm && iterationsSoFar != maxIterations) writeToScreen("Restarting...\n");
	}while(rNormHigh > endNorm && pTAp > 0);
	
	writeToScreen1("Solved the system by %i iterations\n", iterationsSoFar);
	return sqrt(rNormHigh);
}

template <class T, class TLow>
T rcg(CGsystem<T> *AHigh, CGsystem<TLow> *ALow, const T tol, const TLow tol2, const int maxIt, const T threshold, const int bs = 32)
{
	//rnorm is the square of the 2-norm of the residual.
	T oldNorm, previousNorm, rNormHigh, scaleFactor;
	TLow rNormLow, bestNorm, pTAp, eta, alpha, beta;
	unsigned int iterationsSoFar = 0;
	int maxIterations = maxIt;
	int pTApNegativeCounter = 0;
	T * xHigh, *bHigh;
	TLow *xLow, *bLow;
	//initialize variables:
	T **vecArrayHigh	= AHigh->initVectors(&xHigh, &bHigh, 3);
	TLow **vecArrayLow	= ALow->initVectors(&xLow, &bLow, 4);
	T *rHigh	= vecArrayHigh[0];
	T *bestxHigh = vecArrayHigh[1];
	T *tempHigh = vecArrayHigh[2];
	TLow *p			= vecArrayLow[0];
	TLow *rLow	= vecArrayLow[1];
	TLow *Ap		= vecArrayLow[2];
	TLow *bestx	= vecArrayLow[3];
	
	AHigh->scale(0.0, xHigh);						// x = 0
	AHigh->copy(bHigh, &rHigh);				// r = b;
	AHigh->matrixVectorProduct(xHigh, tempHigh); 	// tempHigh = Ax
	AHigh->axpy(-1, tempHigh, rHigh);						// r = b - Ax
	rNormHigh = AHigh->vectorVectorProduct(rHigh, rHigh);
	
	T endNorm = rNormHigh*tol*tol;
	writeToScreen1("endNorm: %g\n", sqrt(endNorm));
	writeToScreen1("rNormHigh: %g\n", sqrt(rNormHigh));
	do{
		previousNorm = rNormHigh;
		scaleFactor = sqrt(rNormHigh);
		AHigh->scale(1.0 / scaleFactor, rHigh);
		copy_kernel<T, TLow, false><<<(ALow->getN() -1) / bs + 1 , bs>>>(rHigh, ALow->getN(), rLow);
		ALow->copy(rLow, &p);
		ALow->scale(0.0, xLow);
		rNormLow = ALow->vectorVectorProduct(rLow, rLow);
//		rNormLow = cublasSdot (ALow->getN(), (float *)rLow, 1, (float*)rLow, 1);
		bestNorm = rNormLow;
		if (threshold * endNorm / rNormHigh > tol2*tol2)
			eta = threshold * endNorm / rNormHigh;
		else
			eta = rNormLow*tol2*tol2;
		unsigned int k;
		writeToScreen1("eta: %g\n", sqrt(eta));
		for (k=0; (k < ALow->getN()) && (k + iterationsSoFar < maxIterations) && (rNormLow > eta); ++k){
			ALow->matrixVectorProduct(p, Ap); //Ap = A*p
			pTAp = ALow->vectorVectorProduct(p, Ap);
//			pTAp = cublasSdot (ALow->getN(), (float *)p, 1, (float*)Ap, 1);
			//if pTAp == 0, we have a lucky breakdown, if pTAp < 0 something is wrong...
			if (pTAp <= (T)0.0) break;
			alpha = rNormLow / pTAp;
			if (alpha == (T)0.0) break;
			//saxpy calculating x += alpha*p and r -= alpha*Ap:
			ALow->axpy(-alpha, Ap, rLow);
			oldNorm = rNormLow;
			rNormLow = ALow->vectorVectorProduct(rLow,rLow);
//			rNormLow = cublasSdot (ALow->getN(), (float *)rLow, 1, (float*)rLow, 1);
			if (rNormLow > oldNorm && oldNorm < bestNorm){
				ALow->copy(xLow, &bestx);
				bestNorm = oldNorm;
			}
			ALow->axpy(alpha, p, xLow);
			beta = rNormLow / oldNorm;
			//saxpy for p = beta * p + r.
			ALow->scale(beta, p);
			ALow->axpy(1.0, rLow, p);
			if (k % 10000 == 0){
/*				if (sizeof(T) == sizeof(float)){
					T crnorm = cublasSdot (A->getN(), (float *)r, 1, (float*)r, 1);
					printf("difference %g\n", crnorm - rnorm);
				}
				if (sizeof(T) == sizeof(double)){
					T crnorm = cublasDdot (A->getN(), (double *)r, 1, (double*)r, 1);
					printf("difference %g\n", crnorm - rnorm);
				}
*/				writeToScreen1("IterationsSoFar = %i\t", k + iterationsSoFar);
				writeToScreen1("norm: %g\t", sqrt(oldNorm));
				writeToScreen1("best norm so far: %g\t", sqrt(bestNorm));
				writeToScreen1("desired norm: %g\n", sqrt(eta));
			}
		}
		iterationsSoFar += k;
		if (rNormLow > bestNorm){
			ALow->copy(bestx, &xLow);
			rNormLow = bestNorm;
		}
		writeToScreen1("rNormLow: %g\n", sqrt(bestNorm));
		/* Update x by the xLow: xHigh = xHigh + ||xHigh||*xLow */
		AHigh->copy(xHigh, &bestxHigh);
		copy_kernel<TLow, T, false><<<(AHigh->getN() -1) / bs + 1 , bs>>>(xLow, AHigh->getN(), tempHigh);		
		AHigh->axpy(scaleFactor, tempHigh, xHigh);
		
		AHigh->matrixVectorProduct(xHigh, tempHigh); 	// p = Ax
		AHigh->copy(bHigh, &rHigh);								// r = b
		AHigh->axpy(-1, tempHigh, rHigh);						// r = b - Ax
		rNormHigh = AHigh->vectorVectorProduct(rHigh, rHigh);
		writeToScreen1("The new norm is: %g\n", sqrt(rNormHigh));
		if (pTAp == 0) return sqrt(rNormHigh);
		if (rNormHigh > endNorm && rNormHigh >= previousNorm){
			writeToScreen("Restart led to the same or worse result\n");
			AHigh->copy(bestxHigh, &xHigh);
			return sqrt(previousNorm);
		}
		if (iterationsSoFar >= maxIterations){
			writeToScreen("Unable to convert to the desired precition, \nbecause the maximum number of iterations was reached\n");
			writeToScreen1("norm: %g\n", sqrt(rNormHigh));
			writeToScreen1("number of iterations: %i\n", iterationsSoFar);
			return sqrt(rNormHigh);
		}
		if (alpha == (T)0.0 || beta == (T)0.0 || pTAp < (T)0.0){
			writeToScreen("Unable to convert to the desired precition, because the iteration stagnated: ");
			if (alpha == (T)0.0) writeToScreen("alpha became zero\n");
			else if (pTAp < (T)0.0){
				pTApNegativeCounter++;
				if (rNormLow > bestNorm){
					ALow->copy(bestx, &xLow);
					rNormLow = bestNorm;
				}
				writeToScreen1("pTAp = %g\n", pTAp);
				copy_kernel<TLow, T, false><<<(AHigh->getN() -1) / bs + 1 , bs>>>(xLow, AHigh->getN(), tempHigh);		
				AHigh->axpy(scaleFactor, tempHigh, xHigh);
				AHigh->matrixVectorProduct(xHigh, tempHigh); 	// p = Ax
				AHigh->copy(bHigh, &rHigh);								// r = b
				AHigh->axpy(-1, tempHigh, rHigh);						// r = b - Ax
				rNormHigh = AHigh->vectorVectorProduct(rHigh, rHigh);
				writeToScreen1("The norm has become %g\n", rNormHigh);
				break;
			}
			else	writeToScreen("beta became zero\n");
			return sqrt(rNormHigh);
		}
		if (rNormHigh > endNorm && iterationsSoFar != maxIterations) writeToScreen("Restarting...\n");
	}while(rNormHigh > endNorm && (pTApNegativeCounter < 5)); //&& pTAp > 0);
	
	writeToScreen1("Solved the system by %i iterations\n", iterationsSoFar);
	return sqrt(rNormHigh);
}
#endif
